extern crate midir;
extern crate pancurses;
extern crate time;
extern crate timer;

use std::error::Error;
use std::thread;
use std::time::{Duration, Instant};
use std::sync::{Arc, Mutex};
use std::sync::atomic::{AtomicUsize, Ordering};
use std::sync::mpsc::channel;

use midir::os::unix::{VirtualInput, VirtualOutput};

use pancurses::{initscr, endwin, noecho, Input};

#[derive(Clone, Debug)]
enum MidiMessage {
    NoteOn(u8, u8, u8),
    NoteOff(u8, u8, u8),
    PitchBend(u8, i16),
    Other(Vec<u8>),
}

impl MidiMessage {
    fn from_slice(s: &[u8]) -> Option<MidiMessage> {
        if s.len() == 0 {
            return None;
        }
        let channel = s[0] & 0x0f;
        match s[0] & 0xf0 {
            0x80 => {
                if s.len() == 3 {
                    Some(MidiMessage::NoteOff(channel, s[1], s[2]))
                } else {
                    None
                }
            },
            0x90 => {
                if s.len() == 3 {
                    Some(MidiMessage::NoteOn(channel, s[1], s[2]))
                } else {
                    None
                }
            },
            0xE0 => {
                if s.len() == 3 {
                    let lsb = s[2] as i16;
                    let msb = s[1] as i16;
                    Some(MidiMessage::PitchBend(channel, (msb << 8) | lsb))
                } else {
                    None
                }
            },
            _ => None
        }.or_else(|| Some(MidiMessage::Other(s.to_vec())))
    }
}


#[derive(Clone, Copy, Debug)]
enum LooperState {
    Idle,
    Recording,
    Playing
}

struct LooperInputCallback {
    output: Arc<Mutex<midir::MidiOutputConnection>>,
    state: Arc<AtomicUsize>,
    outer_start: Arc<Mutex<Instant>>,
    ts_offset: u64,
    msg_buf: Arc<Mutex<Vec<(u64, Vec<u8>)>>>,
    log_msgs: Box<dyn FnMut(u64, Vec<u8>) + Send>,
}

struct Looper {
    virtual_input: Arc<midir::MidiInputConnection<LooperInputCallback>>,
    virtual_output: Arc<Mutex<midir::MidiOutputConnection>>,

    cur_state: Arc<AtomicUsize>,
    // (time in milliseconds, msg)
    msg_buf: Arc<Mutex<Vec<(u64, Vec<u8>)>>>,
    // for either playback or recording
    start_time: Arc<Mutex<Instant>>,
    recording_length: u64,
    playback_structs: Option<timer::Guard>,
    playback_timer: timer::Timer,
    playback_cb: Arc<Mutex<dyn FnMut(u64, MidiMessage) + Send + 'static>>,
}

impl Looper {
    const IDLE: usize = 0;
    const RECORDING: usize = 1;
    const PLAYBACK: usize = 2;

    fn new<F, G>(mut record_cb: F, playback_cb: G) -> Result<Looper, Box<dyn Error>> 
        where F: 'static + FnMut(u64, MidiMessage) + Send,
                G: 'static + FnMut(u64, MidiMessage) + Send
    {
        let input = midir::MidiInput::new("midi looper")?;
        let output = midir::MidiOutput::new("midi looper")?;

        let cur_state = Arc::new(AtomicUsize::new(Looper::IDLE));
        let msg_buf = Arc::new(Mutex::new(Vec::new()));
        let start_time = Arc::new(Mutex::new(Instant::now()));

        let output_unboxed = output.create_virtual("looper output")?;
        let virtual_output = Arc::new(Mutex::new(output_unboxed));

        let virtual_input = input.create_virtual(
            "looper input",
            |ts, data, me| {
                {
                    let mut out = me.output.lock().unwrap();
                    out.send(data).unwrap();
                }
                let ts_millis = ts/1000;
                if me.state.load(Ordering::Acquire) == Looper::RECORDING {
                    let msg_time = Instant::now();

                    let mut buf = me.msg_buf.lock().unwrap();

                    // new recording session; find the new offset
                    if buf.len() == 0 {
                        let record_start = me.outer_start.lock().unwrap();
                        let time_since_start = msg_time.duration_since(*record_start);
                        // let's assume the user pressed their first note less than
                        // 2^64 milliseconds after starting recording
                        let time_millis = time_since_start.as_millis() as u64;
                        me.ts_offset = ts_millis - time_millis;
                    }
                    buf.push((ts_millis - me.ts_offset, data.to_vec()));
                }
                (me.log_msgs)(ts_millis - me.ts_offset, data.to_vec());
            }, LooperInputCallback {
                output: virtual_output.clone(),
                state: cur_state.clone(),
                outer_start: start_time.clone(),
                ts_offset: 0,
                msg_buf: msg_buf.clone(),
                log_msgs: Box::new(move |ts, msg| {
                    if let Some(msg) = MidiMessage::from_slice(msg.as_slice()) {
                        record_cb(ts, msg);
                    }
                }),
            })?;

        Ok(Looper {
            virtual_input: Arc::new(virtual_input),
            virtual_output,

            cur_state: cur_state.clone(),
            msg_buf: msg_buf.clone(),
            start_time: start_time.clone(),
            playback_structs: None,
            playback_timer: timer::Timer::new(),
            playback_cb: Arc::new(Mutex::new(playback_cb)),
            recording_length: 0,
        })
    }

    fn toggle_state(&mut self) -> Result<(), Box<dyn Error>>{
        let now = Instant::now();
        let next_state = (self.cur_state.load(Ordering::Acquire) + 1) % 3;
        match next_state {
            Looper::RECORDING => {
                let mut buf = self.msg_buf.lock().unwrap();
                let mut start_time = self.start_time.lock().unwrap();
                buf.clear();
                *start_time = now;
            },
            Looper::PLAYBACK => {
                let now = Instant::now();
                if self.msg_buf.lock().unwrap().len() == 0 {
                    // empty buffer, no messages to send
                    return Ok(());
                }
                let mut start = self.start_time.lock().unwrap();
                self.recording_length = now.duration_since(*start).as_millis() as u64;
                *start = now;

                // setup/start playback thread
                let mut cur_tick: u64 = 0;
                let mut cur_idx = 0;
                let playback_buf = self.msg_buf.clone();
                let record_len = self.recording_length;
                let output_mutex = self.virtual_output.clone();
                let playback_cb = self.playback_cb.clone();
                let playback_guard = self.playback_timer.schedule_repeating(
                    time::Duration::milliseconds(1),
                    move || {
                        // TODO optimize this so locks happen after
                        // the messages are sent; this would remove
                        // timing jitter caused by lock contention
                        let buf = playback_buf.lock().unwrap();
                        let (ref ts1, ref msg1) = &buf[cur_idx];
                        let mut ts = ts1;
                        let mut msg = msg1;
                        if *ts == cur_tick {
                            let mut output = output_mutex.lock().unwrap();
                            while *ts == cur_tick {
                                output.send(msg.as_slice()).unwrap();
                                if let Some(msg) = MidiMessage::from_slice(msg) {
                                    let mut cb = playback_cb.lock().unwrap();
                                    (&mut *cb)(*ts, msg);
                                }
                                cur_idx += 1;
                                if cur_idx == buf.len() {
                                    cur_idx = 0;
                                    break;
                                }
                                let (ts_next, msg_next) = &buf[cur_idx];
                                ts = ts_next;
                                msg = msg_next;
                            }
                        }
                        cur_tick += 1;
                        if cur_tick >= record_len {
                            cur_tick = 0;
                            cur_idx = 0;
                        }
                    });
                self.playback_structs = Some(playback_guard);
            },
            Looper::IDLE => {
                self.playback_structs = None;
            },
            _ => ()
        }
        self.cur_state.store(next_state, Ordering::Release);
        Ok(())
    }

    fn state(&self) -> LooperState {
        let cur_state = self.cur_state.load(Ordering::Acquire);
        match cur_state {
            0 => LooperState::Idle,
            1 => LooperState::Recording,
            2 => LooperState::Playing,
            _ => unreachable!("invalid looper state reached {}", cur_state)
        }
    }
}

fn main() {
    let window = initscr();

    window.printw("============================\n");
    window.printw("======== MIDI LOOPER =======\n");
    window.printw("============================\n");
    window.printw("\n");
    window.printw("IDLE\n");
    window.printw("\n");
    window.printw("\n");
    window.printw("\n");
    window.printw("\n");
    window.printw("start/stop = space\n");
    window.printw("save = s\n");
    window.printw("exit = Esc/Ctrl-C\n");
    window.refresh();
    window.keypad(true);
    window.nodelay(true);
    noecho();

    let (tx, rx) = channel();
    let tx_record = tx.clone();
    let tx_playback = tx.clone();
    let mut looper = Looper::new(move |ts, msg| {
        tx_record.send((ts, msg)).unwrap();
    }, move |ts, msg| {
        tx_playback.send((ts, msg)).unwrap();
    }).unwrap();
    // TODO turn off delay, and sleep for 1 millisecond between getch calls
    // TODO call refresh once every 16 frames
    let mut count = 0;
    loop {
        match window.getch() {
            Some(Input::Character('\x1b')) => break,
            Some(Input::Character(c)) => {
                match c {
                    ' ' => {
                        looper.toggle_state().unwrap();
                        window.mvprintw(4, 0, match looper.state() {
                            LooperState::Idle =>        "IDLE     ",
                            LooperState::Recording =>   "RECORDING",
                            LooperState::Playing =>     "PLAYING  "
                        });
                        window.mv(12, 0);
                        window.refresh();
                    },
                    's' => {
                    },
                    _ => ()
                }
            },
            Some(input) => {
                window.addstr(&format!("{:?}", input));
            },
            None => ()
        }
        while let Ok((ts, msg)) = rx.try_recv() {
            window.mvprintw(5, 0, "                                        ");
            window.mvprintw(5, 0, format!("{:?} {}.{:03}", msg, ts/1000, ts%1000));
        }

        count = (count + 1) % 16;
        if count == 0 {
            window.mv(12, 0);
            window.refresh();
        }
        thread::sleep(Duration::from_millis(1));
    }
    endwin();
}
